@@ -35,6 +35,9 @@
 #include <polkit-dbus/polkit-dbus.h>
 #endif
 
+#ifdef HAVE_GIO
+#include "file-monitor.h"
+#endif
 
 #define DBUS_ADDRESS_ENVVAR "DBUS_SESSION_BUS_ADDRESS"
 #define DBUS_INTERFACE_STB "org.freedesktop.SystemToolsBackends"
@@ -44,21 +47,34 @@
 
 #define STB_DISPATCHER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), STB_TYPE_DISPATCHER, StbDispatcherPrivate))
 
+#define DEBUG(d,m...) \
+if (G_UNLIKELY (((StbDispatcherPrivate *) d->_priv)->debug)) \
+  { \
+    g_debug (m); \
+  }
+
+enum {
+  PROP_0,
+  PROP_DEBUG
+};
+
 typedef struct StbDispatcherPrivate   StbDispatcherPrivate;
 typedef struct StbDispatcherAsyncData StbDispatcherAsyncData;
 
 struct StbDispatcherPrivate
 {
   DBusConnection *connection;
-  DBusConnection *session_connection;
-
-  GPid   bus_pid;
-  guint  watch_id;
   gchar *platform;
 
 #ifdef HAVE_POLKIT
   PolKitContext *polkit_context;
 #endif
+
+#ifdef HAVE_GIO
+  StbFileMonitor *file_monitor;
+#endif
+
+  guint debug : 1;
 };
 
 struct StbDispatcherAsyncData
@@ -76,6 +92,17 @@ static GObject* stb_dispatcher_constructor (GType                  type,
 					    guint                  n_construct_properties,
 					    GObjectConstructParam *construct_params);
 
+static void     stb_dispatcher_set_property (GObject      *object,
+					     guint         prop_id,
+					     const GValue *value,
+					     GParamSpec   *pspec);
+static void     stb_dispatcher_get_property (GObject      *object,
+					     guint         prop_id,
+					     GValue       *value,
+					     GParamSpec   *pspec);
+
+static gchar*   get_destination            (DBusMessage *message);
+
 
 G_DEFINE_TYPE (StbDispatcher, stb_dispatcher, G_TYPE_OBJECT);
 
@@ -85,72 +112,19 @@ stb_dispatcher_class_init (StbDispatcherClass *class)
   GObjectClass *object_class = G_OBJECT_CLASS (class);
 
   object_class->constructor = stb_dispatcher_constructor;
-  object_class->finalize    = stb_dispatcher_finalize;
+  object_class->set_property = stb_dispatcher_set_property;
+  object_class->get_property = stb_dispatcher_get_property;
+  object_class->finalize = stb_dispatcher_finalize;
 
+  g_object_class_install_property (object_class,
+				   PROP_DEBUG,
+				   g_param_spec_boolean ("debug", "", "",
+							 FALSE,
+							 G_PARAM_READWRITE));
   g_type_class_add_private (object_class,
 			    sizeof (StbDispatcherPrivate));
 }
 
-static void
-on_bus_term (GPid     pid,
-	     gint     status,
-	     gpointer data)
-{
-  StbDispatcher *dispatcher;
-  StbDispatcherPrivate *priv;
-
-  dispatcher = STB_DISPATCHER (data);
-  priv = dispatcher->_priv;
-
-  g_spawn_close_pid (priv->bus_pid);
-  priv->bus_pid = 0;
-
-  /* if the bus dies, we screwed up */
-  g_critical ("Can't live without bus.");
-  g_assert_not_reached ();
-}
-
-static void
-setup_private_bus (StbDispatcher *dispatcher)
-{
-  StbDispatcherPrivate *priv;
-  DBusError error;
-
-  priv = STB_DISPATCHER_GET_PRIVATE (dispatcher);
-  dbus_error_init (&error);
-
-  if (!priv->bus_pid)
-    {
-      /* spawn private bus */
-      static gchar *argv[] = { "dbus-daemon", "--session", "--print-address", "--nofork", NULL };
-      gint output_fd, size;
-      gchar *envvar;
-      gchar str[300] = { 0, };
-
-      if (!g_spawn_async_with_pipes (NULL, argv, NULL,
-				     G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD,
-				     NULL, NULL, &priv->bus_pid,
-				     NULL, &output_fd, NULL, NULL))
-	return;
-
-      priv->watch_id = g_child_watch_add (priv->bus_pid, on_bus_term, dispatcher);
-      size = read (output_fd, str, sizeof (str));
-      str[size - 1] = '\0';
-
-      envvar = g_strdup_printf (DBUS_ADDRESS_ENVVAR "=%s", str);
-      putenv (envvar);
-
-      /* get a connection with the newly created bus */
-      priv->session_connection = dbus_bus_get (DBUS_BUS_SESSION, &error);
-
-      if (dbus_error_is_set (&error))
-	g_critical (error.message);
-
-      dbus_connection_set_exit_on_disconnect (priv->session_connection, FALSE);
-      dbus_connection_setup_with_g_main (priv->session_connection, NULL);
-    }
-}
-
 static void
 async_data_free (StbDispatcherAsyncData *data)
 {
@@ -174,6 +148,104 @@ retrieve_platform (DBusMessage *message)
       return NULL;
 }
 
+#ifdef HAVE_GIO
+static void
+dispatch_file_list (DBusPendingCall *pending_call,
+		    gpointer         data)
+{
+  StbDispatcher *dispatcher;
+  StbDispatcherPrivate *priv;
+  DBusMessage *reply;
+  DBusError error;
+  gchar **files;
+  gint n_files;
+
+  dispatcher = STB_DISPATCHER (data);
+  priv = STB_DISPATCHER_GET_PRIVATE (dispatcher);
+  reply = dbus_pending_call_steal_reply (pending_call);
+  dbus_error_init (&error);
+
+  if (dbus_set_error_from_message (&error, reply))
+    {
+      g_critical (error.message);
+      dbus_error_free (&error);
+      dbus_message_unref (reply);
+      return;
+    }
+
+  if (dbus_message_get_args (reply, &error,
+			     DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &files, &n_files,
+			     DBUS_TYPE_INVALID))
+    {
+      stb_file_monitor_add_files (priv->file_monitor,
+				  dbus_message_get_path (reply),
+				  (const gchar **) files);
+      dbus_free_string_array (files);
+    }
+  else
+    {
+      g_critical (error.message);
+    }
+
+  dbus_message_unref (reply);
+}
+
+static void
+query_file_list (StbDispatcher *dispatcher,
+		 DBusMessage   *message)
+{
+  StbDispatcherPrivate *priv;
+  DBusPendingCall *pending_call;
+  DBusMessage *file_message;
+  gchar *destination;
+
+  priv = STB_DISPATCHER_GET_PRIVATE (dispatcher);
+  destination = get_destination (message);
+
+  if (G_UNLIKELY (!destination))
+    return;
+
+  if (stb_file_monitor_is_object_handled (priv->file_monitor, destination))
+    {
+      g_free (destination);
+      return;
+    }
+
+  file_message = dbus_message_new_method_call (destination,
+					       dbus_message_get_path (message),
+					       DBUS_INTERFACE_STB,
+					       "getFiles");
+
+  dbus_connection_send_with_reply (priv->connection, file_message, &pending_call, -1);
+
+  if (pending_call)
+    {
+      dbus_pending_call_set_notify (pending_call,
+				    dispatch_file_list,
+				    g_object_ref (dispatcher),
+				    (DBusFreeFunction) g_object_unref);
+      dbus_pending_call_unref (pending_call);
+    }
+
+  dbus_message_unref (file_message);
+  g_free (destination);
+}
+
+static void
+object_changed_cb (StbDispatcher *dispatcher,
+		   const gchar   *object_path)
+{
+  StbDispatcherPrivate *priv;
+  DBusMessage *signal;
+
+  priv = STB_DISPATCHER_GET_PRIVATE (dispatcher);
+  signal = dbus_message_new_signal (object_path, DBUS_INTERFACE_STB, "changed");
+
+  dbus_connection_send (priv->connection, signal, NULL);
+  dbus_message_unref (signal);
+}
+#endif /* HAVE_GIO */
+
 static void
 dispatch_reply (DBusPendingCall *pending_call,
 		gpointer         data)
@@ -181,15 +253,34 @@ dispatch_reply (DBusPendingCall *pending_call,
   StbDispatcherPrivate *priv;
   DBusMessage *reply;
   StbDispatcherAsyncData *async_data;
+  DBusError error;
 
   reply = dbus_pending_call_steal_reply (pending_call);
   async_data = (StbDispatcherAsyncData *) data;
   priv = async_data->dispatcher->_priv;
+  dbus_error_init (&error);
+
+  DEBUG (async_data->dispatcher, "sending reply from: %s", dbus_message_get_path (reply));
+
+  if (dbus_set_error_from_message (&error, reply))
+    {
+      g_warning (error.message);
+      dbus_error_free (&error);
+    }
 
-  /* get the platform if necessary */
   if (dbus_message_has_interface (reply, DBUS_INTERFACE_STB_PLATFORM) &&
       dbus_message_has_member (reply, "getPlatform") && !priv->platform)
-    priv->platform = retrieve_platform (reply);
+    {
+      /* get the platform if necessary */
+      priv->platform = retrieve_platform (reply);
+    }
+#ifdef HAVE_GIO
+  else if (dbus_message_has_interface (reply, DBUS_INTERFACE_STB))
+    {
+      /* monitor configuration files */
+      query_file_list (async_data->dispatcher, reply);
+    }
+#endif /* HAVE_GIO */
 
   /* send the reply back */
   dbus_message_set_destination (reply, async_data->destination);
@@ -301,11 +392,13 @@ dispatch_stb_message (StbDispatcher *dispatcher,
       return;
     }
 
+  DEBUG (dispatcher, "dispatching message to: %s", dbus_message_get_path (message));
+
   copy = dbus_message_copy (message);
 
   /* forward the message to the corresponding service */
   dbus_message_set_destination (copy, destination);
-  dbus_connection_send_with_reply (priv->session_connection, copy, &pending_call, -1);
+  dbus_connection_send_with_reply (priv->connection, copy, &pending_call, -1);
 
   if (pending_call)
     {
@@ -332,6 +425,8 @@ return_error (StbDispatcher *dispatcher,
 
   priv = STB_DISPATCHER_GET_PRIVATE (dispatcher);
 
+  DEBUG (dispatcher, "sending error %s from: %s", error_name, dbus_message_get_path (message));
+
   reply = dbus_message_new_error (message, error_name,
 				  "No permissions to perform the task.");
   dbus_connection_send (priv->connection, reply, NULL);
@@ -518,17 +613,56 @@ stb_dispatcher_init (StbDispatcher *dispatcher)
   priv = STB_DISPATCHER_GET_PRIVATE (dispatcher);
   dispatcher->_priv = priv;
 
-  setup_private_bus (dispatcher);
   setup_connection (dispatcher);
 
-  /* we're screwed if we don't have these */
-  g_assert (priv->session_connection != NULL);
+  /* we're screwed if we don't have this */
   g_assert (priv->connection != NULL);
 
 #ifdef HAVE_POLKIT
   priv->polkit_context = polkit_context_new ();
   polkit_context_init (priv->polkit_context, NULL);
 #endif
+
+#ifdef HAVE_GIO
+  priv->file_monitor = stb_file_monitor_new ();
+
+  g_signal_connect_swapped (priv->file_monitor, "object_changed",
+			    G_CALLBACK (object_changed_cb), dispatcher);
+#endif
+}
+
+static void
+stb_dispatcher_set_property (GObject      *object,
+			     guint         prop_id,
+			     const GValue *value,
+			     GParamSpec   *pspec)
+{
+  switch (prop_id)
+    {
+    case PROP_DEBUG:
+      stb_dispatcher_set_debug (STB_DISPATCHER (object),
+				g_value_get_boolean (value));
+      break;
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+    }
+}
+
+static void
+stb_dispatcher_get_property (GObject      *object,
+			     guint         prop_id,
+			     GValue       *value,
+			     GParamSpec   *pspec)
+{
+  switch (prop_id)
+    {
+    case PROP_DEBUG:
+      g_value_set_boolean (value,
+			   stb_dispatcher_get_debug (STB_DISPATCHER (object)));
+      break;
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+    }
 }
 
 static void
@@ -538,22 +672,15 @@ stb_dispatcher_finalize (GObject *object)
 
   priv = STB_DISPATCHER_GET_PRIVATE (object);
 
-  dbus_connection_unref (priv->session_connection);
   dbus_connection_unref (priv->connection);
 
 #ifdef HAVE_POLKIT
   polkit_context_unref  (priv->polkit_context);
 #endif
 
-  /* terminate the private bus */
-  if (priv->bus_pid)
-    {
-      g_source_remove (priv->watch_id);
-      priv->watch_id = 0;
-
-      kill (priv->bus_pid, SIGTERM);
-      priv->bus_pid = 0;
-    }
+#ifdef HAVE_GIO
+  g_object_unref (priv->file_monitor);
+#endif
 
   g_free (priv->platform);
 
@@ -579,3 +706,27 @@ stb_dispatcher_get (void)
 {
   return g_object_new (STB_TYPE_DISPATCHER, NULL);
 }
+
+void
+stb_dispatcher_set_debug (StbDispatcher *dispatcher,
+			  gboolean       debug)
+{
+  StbDispatcherPrivate *priv;
+
+  g_return_if_fail (STB_IS_DISPATCHER (dispatcher));
+
+  priv = STB_DISPATCHER_GET_PRIVATE (dispatcher);
+  priv->debug = debug;
+  g_object_notify (G_OBJECT (dispatcher), "debug");
+}
+
+gboolean
+stb_dispatcher_get_debug (StbDispatcher *dispatcher)
+{
+  StbDispatcherPrivate *priv;
+
+  g_return_val_if_fail (STB_IS_DISPATCHER (dispatcher), FALSE);
+
+  priv = STB_DISPATCHER_GET_PRIVATE (dispatcher);
+  return priv->debug;
+}
